////////////////////////////////////////////////////////////////
// jsInvaders - Copyright (c) 2004 Oliver Wilkinson
//
// This software is provided 'as-is', without any express or implied warranty. 
// In no event will the authors be held liable for any damages arising from the use
// of this software.
//
// Permission is granted to anyone to use this software for any purpose, including 
// commercial applications, and to alter it and redistribute it freely, subject 
// to the following restrictions:
//
// 1. The origin of this software must not be misrepresented; you must not claim that you 
// wrote the original software. If you use this software in a product, an acknowledgment 
// in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such, and must not be 
// misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//////////////////////////////////////////////////////////////////
//
// jsInvaders is released under the zlib/png license, please see
// <http://opensource.org/licenses/zlib-license.php> for more details
//
// jsInvaders is the accompanying code to the article
// "Creating moddable games with XML and Scripting"
// by Oliver Wilkinson
//
////////////////////////////////////////////////////////////////////

#include "jsiGame.h"


namespace evolutionalcouk
{
	namespace jsInvaders
	{
		
		jsiGame::jsiGame()
		{	
		}

		jsiGame::~jsiGame()
		{
		}


		///////////////////////////////////////////////////////////////////
		// Method:		initialise
		///////////////////////////////////////////////////////////////////
		// Access:		public
		// Prototype:	int initialise()
		//
		// Remarks:		Initialises the game by creating the render window
		//				Calls loadXmlLevels to load the game data
		//
		//
		// Parameters:	None
		//
		// Returns:		0 = fail
		//
		// History:		
		//
		//
		////////////////////////////////////////////////////////////////////
		int jsiGame::initialise()
		{
			// create an opengl render window for rendering our play area ;)
			new jsiRenderWindow(&m_gameData);
			if (!jsiRenderWindow::getSingleton().createWindow(640, 480, 32, false))
			{
				// Window failed to create
				jsiLog::getSingleton().writeError( "jsiGame::createWindow() - failed to create game window" );
				shutdown();
				return 0;
			}

			m_gameIsRunning = true;

			return reset();
		}
		///////////////////////////////////////////////////////////////////
		// Method:		update
		///////////////////////////////////////////////////////////////////
		// Access:		public
		// Prototype:	int update()
		//
		// Remarks:		Updated once per frame. Updates the entire game and
		//				calls for a render of the scene
		//
		//
		// Parameters:	None
		//
		// Returns:		1 > for success
		//
		// History:		16/06/2004 - Created skeleton function
		//				18/06/2004 - Added the updating of the renderer
		//				20/06/2004 - Added a isRunning check to see if the game
		//				has requested a shutdown. If false, kill the game
		//				20/06/2004 - Added the ability to reload the level
		//				by pressing the 'r' key
		////////////////////////////////////////////////////////////////////
		int jsiGame::update()
		{
			// First check if the game is running, if not call shut down
			if (!m_gameIsRunning)
			{
				return 0;
			}
			// Try to render the scene
			if (!jsiRenderWindow::getSingleton().renderScene())
			{
				// if the scee rendering failed, we should return 0
				jsiLog::getSingleton().writeError( "jsiRenderWindow::renderScene() - Failed to update scene" );
				return 0;
			}

			// Check to see if a reload is needed (check kepress 'r')
			if (m_gameData.keys[ 82 ] == true)
			{
				// call reset
				return reset();
			}

			return 1;
		}
		
		///////////////////////////////////////////////////////////////////
		// Method:		reset
		///////////////////////////////////////////////////////////////////
		// Access:		public
		// Prototype:	int reset()
		//
		// Remarks:		Updated once per frame. Updates the entire game and
		//				calls for a render of the scene
		//
		//
		// Parameters:	None
		//
		// Returns:		1 > for success
		//
		// History:		08/08/2004 - Altered build options so that game runs in "dist/"
		//							 The parameter passed to loadXmlLevels now reflects this
		//
		////////////////////////////////////////////////////////////////////
		int jsiGame::reset()
		{
			// Clear the game for levels
			clearLevels();

			if (!loadXmlLevels("level.xml"))
			{
				// level failed to load, report an error
				jsiLog::getSingleton().writeError( "jsiGame::loadXmlLevels - Failed to load the level file" );
				shutdown();
				return 0;
			}

			// Level should be loaded ok here, fill in the shared gamedata structure
			//
			// get first level
			m_gameData.currentLevel = m_levels.get(0);
			// set player stats
			m_gameData.playerHealth = 100;
			m_gameData.playerLives = 3;
			m_gameData.playerScore = 0;
			ZeroMemory( m_gameData.keys, 256 * sizeof(bool) );

			return 1;
		};

		int jsiGame::shutdown()
		{
			m_gameIsRunning = false;

			// Clear out all our levels
			clearLevels();

			// close down our render window and delete it
			if (jsiRenderWindow::getSingletonPtr())
			{
				jsiRenderWindow::getSingleton().destroyWindow();

				delete jsiRenderWindow::getSingletonPtr();
			}

			return 0;
		}
		
		///////////////////////////////////////////////////////////////////
		// Method:		createLevel
		///////////////////////////////////////////////////////////////////
		// Access:		public
		// Prototype:	jsiLevel* createLevel()
		//
		// Remarks:		Creates a level and associates it with the game
		//
		//
		// Parameters:	None
		//
		// Returns:		A pointer to the new level
		//
		// History:		17/06/2004 - Created initial prototype
		//				17/06/2004 - Swapped out original for simply factory call
		//
		//
		////////////////////////////////////////////////////////////////////
		jsiLevel *jsiGame::createLevel()
		{
			return m_levels.create();
		}

		void jsiGame::clearLevels()
		{
			m_levels.clear();
		}

		///////////////////////////////////////////////////////////////////
		// Method:		loadXmlLevels
		///////////////////////////////////////////////////////////////////
		// Access:		public
		// Prototype:	int loadXmlLevels(const char *filename)
		//
		// Remarks:		Loads and parses an xml level file
		//
		//
		// Parameters:	A char * that contains the filename
		//
		// Returns:		>0 = Success
		//
		// History:		17/06/2004 - Created
		//
		//
		////////////////////////////////////////////////////////////////////
		int jsiGame::loadXmlLevels(const char *filename)
		{
			// Load our xml document into a newly created TiXmlDocument class
			// TinyXml will build an internal representation of the document within itself
			// This is stored as a hierarchy of linked-list elements, you would
			// traverse this list as you would, say, a file tree
			TiXmlDocument	*xmlDoc = new TiXmlDocument();
			if (!xmlDoc->LoadFile(filename))
			{
				// Fatal error, cannot load
				jsiLog::getSingleton().writeError( "jsiGame::loadXmlLevels() - Failed to load Xml document" );
				return 0;
			}

			// Define some Xml elements to parse
			TiXmlElement *xGame = 0;
			// Try and grab the document game element from the root
			// TinyXml can search the document tree that it built up internally
			// after loading it.
			xGame = xmlDoc->FirstChildElement("invadersgame");

			if (!xGame)
			{
				// The game element couldn't be found, we have to say goodbye!
				// clear and delete the doc before exiting
				jsiLog::getSingleton().writeError( "jsiGame::loadXmlLevels() - <invadersgame> element not found" );
				xmlDoc->Clear();
				delete xmlDoc;

				return 0;
			}
			//
			// We have found our game, we should be looking for level declarations
			// Remember that Xml files are stored in an hierarchy, so instead of looking
			// for a level in the document root we look in the game element we have just
			// obtained.
			//
			TiXmlElement *xLevel = 0;
			xLevel = xGame->FirstChildElement("level");
			
			if (!xLevel)
			{
				// No level elements were found - this isn't good for our game now!
				// return with an error
				// clear and delete the doc before exiting
				jsiLog::getSingleton().writeError( "jsiGame::loadXmlLevels() - No <level> elements found in Xml document" );
				xmlDoc->Clear();
				delete xmlDoc;
				return 0;
			}

			// TODO: This will only parse a single level so far, this will be changed
			// in later iterations of the code - it can easily be stuck into a loop ;)

			// Because we have a level element, we should create a new level object for
			// our game and pass the xml level description for it to parse
			
			jsiLevel *newLevel = m_levels.create();
			newLevel->parseXmlLevel(xLevel);

			// Clear the internal TinyXml document of all elements (this will ensure that
			// all data is freed up cleanly)
			xmlDoc->Clear();
			// Delete our allocated document and return success ;)
			delete xmlDoc;
			return 1;
		}
		///////////////////////////////////////////////////////////////////
		// Method:		isRunning
		///////////////////////////////////////////////////////////////////
		// Access:		public
		// Prototype:	bool isRunning()
		//
		// Remarks:		Returns the current run status of the game. If this is
		//				false then update should not be called
		//
		// Parameters:	None
		//
		// Returns:		bool
		//
		// History:		20/06/2004 - Created
		//
		//
		////////////////////////////////////////////////////////////////////
		bool jsiGame::isRunning()
		{
			return m_gameIsRunning;
		}

	}
}
