////////////////////////////////////////////////////////////////
// jsInvaders - Copyright (c) 2004 Oliver Wilkinson
//
// This software is provided 'as-is', without any express or implied warranty. 
// In no event will the authors be held liable for any damages arising from the use
// of this software.
//
// Permission is granted to anyone to use this software for any purpose, including 
// commercial applications, and to alter it and redistribute it freely, subject 
// to the following restrictions:
//
// 1. The origin of this software must not be misrepresented; you must not claim that you 
// wrote the original software. If you use this software in a product, an acknowledgment 
// in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such, and must not be 
// misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//////////////////////////////////////////////////////////////////
//
// jsInvaders is released under the zlib/png license, please see
// <http://opensource.org/licenses/zlib-license.php> for more details
//
// jsInvaders is the accompanying code to the article
// "Creating moddable games with XML and Scripting"
// by Oliver Wilkinson
//
////////////////////////////////////////////////////////////////////
//
// Please note that large portions of code in this subsystem is based upon 
// Jeff Molofee's (NeHe) OpenGL tutorials located at <http://nehe.gamedev.net>
//
// Thanks to Jeff for creating such tutorials and making them public, it allows
// small games like this one to be created and shared with the game development community
//
////////////////////////////////////////////////////////////////////

#include "render/jsiRenderWindow.h"

namespace evolutionalcouk
{
	namespace jsInvaders
	{

		jsiRenderWindow::jsiRenderWindow(jsiGameData *gameData)
		{
			m_gameData = gameData;

			// Define the shape of the alien 'model' here for now
			// Simple square 2x2 with centre as 0,0
			m_basicShape.tl = jsiVertex( -1.0f, 1.0f, 0.0f );
			m_basicShape.tr = jsiVertex( 1.0f, 1.0f, 0.0f );
			m_basicShape.br = jsiVertex( 1.0f, -1.0f, 0.0f );
			m_basicShape.bl = jsiVertex( -1.0f,-1.0f, 0.0f );
		}

		///////////////////////////////////////////////////////////////////
		// Method:		createWindow
		///////////////////////////////////////////////////////////////////
		// Access:		public
		// Prototype:	int createWindow(const int width, const int height, const int depth, const bool fullscreen)
		//
		// Remarks:		Creates the window used for the game. Sets up OpenGL handles
		//				and inits device context and pixel format
		//
		// Parameters:	Width, height and colour depth of the window. Fullscreen is unused at present and the function
		//				will fail if it is specified
		//
		// Returns:		0 = fail
		//
		// History:		18/06/2004 - Added to project
		//				20/06/2004 - Documented
		//				20/06/2004 - TODO: Add fullscreen support
		//
		////////////////////////////////////////////////////////////////////
		int jsiRenderWindow::createWindow(const int width, const int height, const int depth, const bool fullscreen)
		{
	
			// TODO 20/06/2004: Add fullscreen support - if fullscreen is specified, function will fail until
			// support is added
			if (fullscreen)
			{
				MessageBox(NULL, "jsiRenderWindow::createWindow() - Fullscreen not supported at present", "ERROR", MB_OK | MB_ICONEXCLAMATION);
				jsiLog::getSingleton().writeError( "jsiRenderWindow::createWindow() - Fullscreen not supported at present" );
				return 0;
			}

			GLuint		pxFmt;						// Holds The Results After Searching For A Match
			WNDCLASS	wc;							// Windows Class Structure
			DWORD		dwExStyle;						// Window Extended Style
			DWORD		dwStyle;						// Window Style
			
			RECT rect;							// Grabs Rectangle Upper Left / Lower Right Values
			rect.left = (long)0;						// Set Left Value To 0
			rect.right = (long)width;						// Set Right Value To Requested Width
			rect.top =  (long)0;							// Set Top Value To 0
			rect.bottom = (long)height;						// Set Bottom Value To Requested Height
			
			m_fullscreen = fullscreen;

			m_hInst				= GetModuleHandle(NULL);			// Grab An Instance For Our Window
			wc.style			= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;		// Redraw On Move, And Own DC For Window
			wc.lpfnWndProc		= (WNDPROC)g_gameWndProc;				// WndProc Handles Messages
			wc.cbClsExtra		= 0;						// No Extra Window Data
			wc.cbWndExtra		= 0;						// No Extra Window Data
			wc.hInstance		= m_hInst;					// Set The Instance
			wc.hIcon			= LoadIcon(NULL, IDI_WINLOGO);			// Load The Default Icon
			wc.hCursor			= LoadCursor(NULL, IDC_ARROW);			// Load The Arrow Pointer
			wc.hbrBackground	= NULL;						// No Background Required For GL
			wc.lpszMenuName		= NULL;						// We Don't Want A Menu
			wc.lpszClassName	= "jsInvaders";					// Set The Class Name

			if (!RegisterClass(&wc))						// Attempt To Register The Window Class
			{
				jsiLog::getSingleton().writeError( "jsiRenderWindow::createWindow() - Failed to register window class" );
				MessageBox(NULL, "jsiRenderWindow::createWindow() - Failed to register window class", "ERROR", MB_OK|MB_ICONEXCLAMATION);
				return 0;							// Exit And Return FALSE
			}

			if (fullscreen)
			{

			}
			else
			{
				dwExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;			// Window Extended Style
				dwStyle = WS_OVERLAPPEDWINDOW;					// Windows Style
			}

			AdjustWindowRectEx(&rect, dwStyle, FALSE, dwExStyle);		// Adjust Window To True Requested Size

			m_hWnd = CreateWindowEx( dwExStyle,				// Extended Style For The Window
					"jsInvaders",				// Class Name
					"jsInvaders",				// Window Title
					WS_CLIPSIBLINGS |			// Required Window Style
					WS_CLIPCHILDREN |			// Required Window Style
					dwStyle,
					0, 0,					// Window Position
					rect.right - rect.left,	// Calculate Adjusted Window Width
					rect.bottom - rect.top,	// Calculate Adjusted Window Height
					(HWND) NULL,					// No Parent Window
					NULL,					// No Menu
					m_hInst,				// Instance
					NULL);					// Don't Pass Anything To WM_CREATE
			
			if (!m_hWnd)
			{
				MessageBox(NULL, "jsiRenderWindow::createWindow() - Cannot create game window", "ERROR", MB_OK|MB_ICONEXCLAMATION);
				jsiLog::getSingleton().writeError( "jsiRenderWindow::createWindow() - Cannot create game window" );
				return 0;							// Return FALSE
			}

			// The window should be created here
			// now obtain some info about the pixel format, etc
			static	PIXELFORMATDESCRIPTOR pfd=					// pfd Tells Windows How We Want Things To Be
				{
					sizeof(PIXELFORMATDESCRIPTOR),					// Size Of This Pixel Format Descriptor
					1,								// Version Number
					PFD_DRAW_TO_WINDOW |						// Format Must Support Window
					PFD_SUPPORT_OPENGL |						// Format Must Support OpenGL
					PFD_DOUBLEBUFFER,						// Must Support Double Buffering
					PFD_TYPE_RGBA,							// Request An RGBA Format
					depth,								// Select Our Color Depth
					0, 0, 0, 0, 0, 0,						// Color Bits Ignored
					0,								// No Alpha Buffer
					0,								// Shift Bit Ignored
					0,								// No Accumulation Buffer
					0, 0, 0, 0,							// Accumulation Bits Ignored
					16,								// 16Bit Z-Buffer (Depth Buffer)
					0,								// No Stencil Buffer
					0,								// No Auxiliary Buffer
					PFD_MAIN_PLANE,							// Main Drawing Layer
					0,								// Reserved
					0, 0, 0								// Layer Masks Ignored
				};

			// Get device context for the window
			if (!(m_hDC = GetDC(m_hWnd)))							// Did We Get A Device Context?
			{
				destroyWindow();
				jsiLog::getSingleton().writeError( "jsiRenderWindow::createWindow() - Unable to create OpenGL device context" );
				MessageBox(NULL, "jsiRenderWindow::createWindow() - Unable to create OpenGL device context"," ERROR",MB_OK|MB_ICONEXCLAMATION);
				return FALSE;							// Return FALSE
			}

			// let windows find a correct pixel format
			if (!(pxFmt = ChoosePixelFormat( m_hDC, &pfd)))				// Did Windows Find A Matching Pixel Format?
			{
				destroyWindow();
				jsiLog::getSingleton().writeError( "jsiRenderWindow::createWindow() - Cannot find a suitable pixel format for desired depth" );
				MessageBox(NULL, "jsiRenderWindow::createWindow() - Cannot find a suitable pixel format for desired depth", "ERROR",MB_OK|MB_ICONEXCLAMATION);
				return FALSE;							// Return FALSE
			}
			// set the pixel format
			if(!SetPixelFormat( m_hDC, pxFmt, &pfd))				// Are We Able To Set The Pixel Format?
			{
				destroyWindow();
				jsiLog::getSingleton().writeError( "jsiRenderWindow::createWindow() - Unable to set pixel format" );
				MessageBox(NULL, "jsiRenderWindow::createWindow() - Unable to set pixel format", "ERROR",MB_OK|MB_ICONEXCLAMATION);
				return FALSE;							// Return FALSE
			}

			if (!(m_hRC = wglCreateContext(m_hDC)))					// Are We Able To Get A Rendering Context?
			{
				destroyWindow();
				jsiLog::getSingleton().writeError( "jsiRenderWindow::createWindow() - Failed to create OpenGL rendering context" );
				MessageBox(NULL, "jsiRenderWindow::createWindow() - Failed to create OpenGL rendering context", "ERROR",MB_OK|MB_ICONEXCLAMATION);
				return FALSE;							// Return FALSE
			}

			if(!wglMakeCurrent(m_hDC,m_hRC))						// Try To Activate The Rendering Context
			{
				destroyWindow();
				jsiLog::getSingleton().writeError( "jsiRenderWindow::createWindow() - Failed to activate OpenGL rendering context" );
				MessageBox(NULL, "jsiRenderWindow::createWindow() - Failed to activate OpenGL rendering context", "ERROR",MB_OK|MB_ICONEXCLAMATION);
				return FALSE;							// Return FALSE
			}

			// Now we need to show the window and set focus to it
			ShowWindow( m_hWnd,SW_SHOW);
			SetForegroundWindow( m_hWnd);
			SetFocus(m_hWnd);
			// refresh the scene for the initial draw
			resizeGLScene(width, height);

			if (!initGL())
			{
				jsiLog::getSingleton().writeError( "jsiRenderWindow::createWindow() - OpenGL initialisation failed" );
				MessageBox(NULL, "jsiRenderWindow::createWindow() - OpenGL initialisation failed", "ERROR",MB_OK|MB_ICONEXCLAMATION);
				return 0;
			}

			return 1;
		}

		///////////////////////////////////////////////////////////////////
		// Method:		destroyWindow
		///////////////////////////////////////////////////////////////////
		// Access:		public
		// Prototype:	void destroyWindow()
		//
		// Remarks:		Closes the window and releases all handles obtained
		//				for OpenGL use. Called on initialisation failure
		//				or on game exit
		//
		// Parameters:	None
		//
		// Returns:		None
		//
		// History:		18/06/2004 - Added to project
		//				20/06/2004 - Documented
		//				08/08/2004 - Fixed incorrect name bug in the window class which raised a logfile warning
		//
		////////////////////////////////////////////////////////////////////
		void jsiRenderWindow::destroyWindow()
		{
			if (m_hDC)
			{
				// ask opengl to release the render and device context
				if (! wglMakeCurrent(NULL, NULL) )	
				{
					// error
					jsiLog::getSingleton().writeError( "jsiRenderWindow::destroyWindow() - Unable to restore windows device context" );
				}
				// ask opengl to delete the rendering context
				if (!wglDeleteContext(m_hRC))
				{
					// error
					jsiLog::getSingleton().writeError( "jsiRenderWindow::destroyWindow() - Unable to delete OpenGL rendering contextt" );
				}
				m_hDC = 0;

				if ( m_hDC && !ReleaseDC(m_hWnd, m_hDC) )
				{
					// error
					jsiLog::getSingleton().writeError( "jsiRenderWindow::destroyWindow() - Unable to release device context" );
					m_hDC = 0;
				}

				// destroy our window, if it exists
				if ( m_hWnd && !DestroyWindow(m_hWnd) )
				{
					// failed :<
					jsiLog::getSingleton().writeError( "jsiRenderWindow::destroyWindow() - Unable to destroy game window" );
					m_hWnd = 0;
				}
				// finally, release our window class
				if ( !UnregisterClass("jsInvaders", m_hInst) )
				{
					// error
					jsiLog::getSingleton().writeWarning( "jsiRenderWindow::destroyWindow() - Unable to release window class" );
					m_hInst = 0;
				}

			}
		}


		///////////////////////////////////////////////////////////////////
		// Method:		g_gameWndProc
		///////////////////////////////////////////////////////////////////
		// Access:		public (static)
		// Prototype:	LRESULT CALLBACK gameWndProc(HWND hwnd,UINT uMsg, WPARAM wParam, LPARAM lParam)
		//
		// Remarks:		Windows needs a static method to call for a windows procedure.
		//				This function acts as a proxy by being static but referring the
		//				messages over to the actual message procedure in the class
		// Parameters:	Window, message and message parameters (WPARAM/LPARAM)
		//
		// Returns:		LRESULT
		//
		// History:		18/06/2004 - Added to project
		//				20/06/2004 - Documented
		//
		//
		////////////////////////////////////////////////////////////////////
		LRESULT CALLBACK g_gameWndProc(HWND hwnd,UINT uMsg, WPARAM wParam, LPARAM lParam)
		{
			return jsiRenderWindow::getSingleton().gameWndProc(hwnd, uMsg, wParam, lParam);
		}

		///////////////////////////////////////////////////////////////////
		// Method:		gameWndProc
		///////////////////////////////////////////////////////////////////
		// Access:		public
		// Prototype:	LRESULT gameWndProc(HWND hwnd,UINT uMsg, WPARAM wParam, LPARAM lParam)
		//
		// Remarks:		The Windows message procedure for the window. Handles all messages
		//				sent from windows, including kepresses and resizing the window
		//
		// Parameters:	Window, message and message parameters (WPARAM/LPARAM)
		//
		// Returns:		LRESULT
		//
		// History:		18/06/2004 - Added to project
		//				20/06/2004 - Documented
		//
		////////////////////////////////////////////////////////////////////
		LRESULT jsiRenderWindow::gameWndProc(HWND hwnd,UINT uMsg, WPARAM wParam, LPARAM lParam)
		{
			switch (uMsg)
			{
			case WM_CREATE:
				{
					// Create window message - nothing special occurs
					return 0;
				}
			case WM_CLOSE:
				{
					// Handles the windows close message, we're going to kill off the App now
					PostQuitMessage(0);
					return 0;
				}
			case WM_SIZE:
				{
					// The window has been resized, let's tell OGL
					// lo = width, hi = height
					resizeGLScene( LOWORD(lParam), HIWORD(lParam) );
					return 0;
				}
			case WM_KEYDOWN:
				{
					m_gameData->keys[wParam] = true;
					return 0;
				}
			case WM_KEYUP:
				{
					m_gameData->keys[wParam] = false;
					return 0;
				}
			default:
				{
					// no handler, return the default proc defined by windows
					return DefWindowProc(hwnd, uMsg, wParam, lParam);
				}
			};

			return 0;
		}
		///////////////////////////////////////////////////////////////////
		// Method:		initGL
		///////////////////////////////////////////////////////////////////
		// Access:		private
		// Prototype:	bool initGL()
		//
		// Remarks:		Initialises OpenGL by setting the shading model
		//				and enabling the depth buffer, etc. Called once after the
		//				window has bee created successfully
		//
		// Parameters:	None
		//
		// Returns:		Always true
		//
		// History:		18/06/2004 - Added
		//				20/06/2004 - Documented
		//
		////////////////////////////////////////////////////////////////////
		bool jsiRenderWindow::initGL()
		{
			glShadeModel(GL_SMOOTH);						// Enables Smooth Shading
			glClearColor(0.0f, 0.0f, 0.0f, 0.0f);			// Black Background
			glClearDepth(1.0f);								// Depth Buffer Setup
			glEnable(GL_DEPTH_TEST);						// Enables Depth Testing
			glDepthFunc(GL_LEQUAL);							// The Type Of Depth Test To Do
			glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);			// Really Nice Perspective Calculations
			return true;
		}
		///////////////////////////////////////////////////////////////////
		// Method:		resizeGLScene
		///////////////////////////////////////////////////////////////////
		// Access:		private
		// Prototype:	GLvoid resizeGLScene(GLsizei width, GLsizei height)
		//
		// Remarks:		Resizes the OpenGL viewport and resets the perspective
		//				when window size has been changed
		//
		//
		// Parameters:	Width and height of the new viewport
		//
		// Returns:		None
		//
		// History:		18/06/2004 - Added to project
		//				20/06/2004 - Documented
		//
		////////////////////////////////////////////////////////////////////
		GLvoid jsiRenderWindow::resizeGLScene(GLsizei width, GLsizei height)
		{
			// protect against a zero-height window?
			if (height==0) { height = 1; }
			// Reset The Current Viewport
			glViewport(0, 0, width, height);
			// Set the projection matrix to identity to clear it ...
			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			// and then calculate The Aspect Ratio Of The Window
			gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 100.0f);
			// now we reset the model view matrix
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
		}

		///////////////////////////////////////////////////////////////////
		// Method:		renderScene
		///////////////////////////////////////////////////////////////////
		// Access:		public
		// Prototype:	bool renderScene(jsiLevel *level)
		//
		// Remarks:		Here's where the level itself is actually rendered
		//				
		//
		//
		// Parameters:	None
		//
		// Returns:		true for success
		//
		// History:		19/06/2004 - Added to project
		//
		//
		////////////////////////////////////////////////////////////////////
		bool jsiRenderWindow::renderScene()
		{
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);			// Clear The Screen And The Depth Buffer
			glLoadIdentity();							// Reset The Current Modelview Matrix
			
			renderCurrentLevel();
			
			SwapBuffers(m_hDC);
			return true;
		}
		///////////////////////////////////////////////////////////////////
		// Method:		renderCurrentLevel
		///////////////////////////////////////////////////////////////////
		// Access:		private
		// Prototype:	bool renderCurrentLevel(jsiLevel *level)
		//
		// Remarks:		Here's where the level itself is actually rendered,
		//				it uses the currentLevel from the jsiGameData structure
		//				passed on creation of the window to render the level
		//
		//
		// Parameters:	None
		//
		// Returns:		true for success
		//
		// History:		19/06/2004 - Added to project
		//
		//
		////////////////////////////////////////////////////////////////////
		bool jsiRenderWindow::renderCurrentLevel()
		{
			glTranslatef( 0.0f, 0.0f, -50.0f);

			size_t nAlien = 0;
			size_t nAlienCount = m_gameData->currentLevel->getAlienCount();

			// Loop through the aliens and render
			for (nAlien = 0; nAlien < nAlienCount; nAlien++)
			{
				jsiAlien *alien = m_gameData->currentLevel->getAlien(nAlien);
				jsiQuad qPos = m_basicShape;
				
				qPos.tr += alien->position;
				qPos.tl += alien->position;
				qPos.bl += alien->position;
				qPos.br += alien->position;

				// select the correct colour
				if (alien->color == JSI_ALIENCOLOR_BLUE)
				{
					glColor3f( 0.0f, 0.0f, 1.0f );
				}
				if (alien->color == JSI_ALIENCOLOR_RED)
				{
					glColor3f( 1.0f, 0.0f, 0.0f );
				}
				if (alien->color == JSI_ALIENCOLOR_GREEN)
				{
					glColor3f( 0.0f, 1.0f, 0.0f );
				}

				renderGLQuad(qPos);
			}

			
			
			return true;
		}
		///////////////////////////////////////////////////////////////////
		// Method:		renderGLQuad
		///////////////////////////////////////////////////////////////////
		// Access:		private
		// Prototype:	void renderGLQuad(const jsiQuad &q)
		//
		// Remarks:		A utility function to simplify rendering quads in OpenGL.
		//				Uses renderGLVertex() to render the vertices in the quad
		//
		//
		// Parameters:	Reference to a jsiQuad structure
		//
		// Returns:		None
		//
		// History:		19/06/2004 - Added
		//				20/06/2004 - Documented
		//
		////////////////////////////////////////////////////////////////////
		void jsiRenderWindow::renderGLQuad(const jsiQuad &q)
		{
			glBegin(GL_QUADS);
				renderGLVertex( q.tl );
				renderGLVertex( q.tr );
				renderGLVertex( q.br );
				renderGLVertex( q.bl );
			glEnd();
		}
		///////////////////////////////////////////////////////////////////
		// Method:		renderGLVertex
		///////////////////////////////////////////////////////////////////
		// Access:		private
		// Prototype:	void renderGLVertex(const jsiVertex &v)
		//
		// Remarks:		A utility function to simplify rendering of OpenGL vertices
		//				from a jsiVertex class.
		//				NOTE: This MUST be called between a glBegin() and glEnd() block
		//
		// Parameters:	Reference to the vertex
		//
		// Returns:		None
		//
		// History:		19/06/2004 - Added
		//				20/06/2004 - Documented
		//
		////////////////////////////////////////////////////////////////////
		void jsiRenderWindow::renderGLVertex(const jsiVertex &v)
		{
			glVertex3f( v.x, v.y, v.z );
		}

	}


}